;****************************************************************
;		     HuSound- Sample Playback			*
;****************************************************************
; The PCM sample directory pointed to by HuSample_DirBase must
; reside at an address other than zero and follow the
; structure given below:
;
; Offset	Contents
; -------	---------
; 0		"P"
; 1		"C"
; 2		"M"
; 3		Number of Entries
; 4+		8-Byte PCM Descriptors:
;  \
;   0			"X"
;   1-4			Start Address (Relative Lo, Hi, Bank, Map)
;   5-7			Stream length (Lo, Mid, Hi)
;
;
; Restrictions
; -------------
; The directory must not cross a bank boundary (at its max size
; of 256 entries, the directory will be 2052 bytes). It's easiest
; just to keep it bank aligned.
;
; PCM streams have no alignment or boundary restrictions, but
; must not exceed a maximum size of 8MB.
;
;
; Notable Functions
; -----------------
; -IRQ_HuSample:
;	The IRQ Handler used for sample playback (tied to TIMER).
;
; -HuSample_PlaySimple:
;	Play the sample whose number is in A over channel zero
; with maximum volume and center panning The channel mode is
; automatically switched to DDA as well.
;
; -HuSample_StopSimple:
;	Stops any samples playing back over channel zero.
;
; -HuSample_Play:
;	Play the sample whose number is in A over the channel
; whose number is in Y. The channel must be switched to DDA
; mode and the panning set beforehand.
;
; -HuSample_Stop:
;	Cease any sample playback on the channel number in Y.

;****************************************************************
; IRQ_HuSample:							*
;	IRQ Handler for HuSound's sample playback, tied to the	*
; TIMER interrupt @ ~6.9KHz, uses the Low OBJ Slot for banking.	*
;****************************************************************
.section "IRQ_HuSample" force
IRQ_HuSample:
	PHA
	PHX
	PHY
	STZ IRQ_STAT		; Ack Timer IRQ

	LDA MCG_LastMap
	PHA
	TMA #MPR4		; Preserve the current MCGenjin Map
	PHA			; and object slot bank selection.

	LDX #(HUSOUND_CHANNELS - 1)
	STZ HuSample_ActiveAcc	; <- Incremented for each active stream
_checkActive
	LDA HuSample_Playing,X
	bne _updateSample
	DEX
	bpl _checkActive
	jmp _checkAllDone

_updateSample
	INC HuSample_ActiveAcc

	LDA HuSample_Src_McMap,X
	LDY SysRegion		; Set the MCGenjin Slot and MPR Bank
	beq _setDirect		; the sample is in...
	TAY
	LDA MCG_SwapTable,Y	; Make sure to reverse the write if
_setDirect			; we're in a non-native region.
	STA MCGENJIN_BANK
	LDA HuSample_Src_Bank,X
	TAM #MPR4

	LDA HuSample_Src_Hi,X
	STA HuSample_CurStream + HI
	LDA HuSample_Src_Lo,X
	STA HuSample_CurStream + LO

	STX PSG_CHANNEL
	LDA (HuSample_CurStream)
	STA PSG_WAVEDAT		; ~Write the new sample~

	DEC HuSample_Rem_Lo,X	; Check to see how much data is left
	bne _stillPlaying	; for the sample playback stream.
	LDA HuSample_Rem_Hi,X
	beq _highDone

	LDA HuSample_Rem_Md,X
	DEA
	STA HuSample_Rem_Md,X
	CMP #$FF
	bne _stillPlaying
	DEC HuSample_Rem_Hi,X
	bra _stillPlaying
_highDone
	LDA HuSample_Rem_Md,X
	DEA
	STA HuSample_Rem_Md,X
	CMP #$FF
	bne _stillPlaying


_streamDone			; -- STREAM IS FINISHED --
	STZ HuSample_Playing,X	; Release the channel, enable writebacks.
	STZ HuSound_noWriteback,X
	LDA #$1F
	STA HuSound_Vol_Last,X

	STZ HuSample_Priority,X ; <- Also clear the priority indicator, so
	DEX			; the SFX and MUS decoders argue properly.
	bpl _checkActive

_stillPlaying			; -- STREAM IS STILL PLAYING --
	INC HuSample_Src_Lo,X	; Calculate the next source address, bank,
	bne _noWrapSrc		; and MCGenjin mapping configuration (phew!)
	INC HuSample_Src_Hi,X
	LDA HuSample_Src_Hi,X
	CMP #>ADDR_OBJ_HI
	bne _noWrapSrc

	LDA #>ADDR_OBJ_LO
	STA HuSample_Src_Hi,X
	INC HuSample_Src_Bank,X
	LDA HuSample_Src_Bank,X
	CMP #(BANK_ROM + MCBANK_PAGES)
	bne _noWrapSrc

	LDA #BANK_ROM
	STA HuSample_Src_Bank,X
	INC HuSample_Src_McMap,X
_noWrapSrc
	DEX
	bmi _checkAllDone
	jmp _checkActive


_checkAllDone
	LDA HuSample_ActiveAcc	; If all of our sample streams have
	bne _samplesNotDone	; finished playback, we can disable
	STZ TIMER_EN		; the timer interrupt.
_samplesNotDone
	PLA			; Otherwise, let's get out of here
	TAM #MPR4		; before we've wasted even more cycles.
	PLA
	STA MCGENJIN_BANK

	PLY
	PLX
	PLA
	rti
.ends




;****************************************************************
; HuSample_PlaySimple:						*
;	Play the PCM stream whose number is in A on channel 0	*
; using maximum volume and center panning.			*
;****************************************************************
.section "HuSample_PlaySimple" force
HuSample_PlaySimple:
	PHY			; We don't need to double up the
	CLY			; range check here, since HuSample_Play
	PHA			; dropping a request and leaving the
				; channel in DDA mode will be recovered
	sei			; from on the next writeback.
	STZ PSG_CHANNEL
	LDA #(PSG_CONTROL_SAMPLE | $1F)
	STA PSG_CONTROL
	LDA #$FF
	STA PSG_PAN
	STA HuSample_Priority	; <- The sample priority is automatically
				; forced to its maximum so no one else can
	PLA			; knock it out of playback except another
	jsr HuSample_Play	; HuSample_PlaySimple call.
	cli

	PLY
	rts
.ends
;****************************************************************
; HuSample_StopSimple:						*
;	Stop PCM Playback (if active) on channel 0.		*
;****************************************************************
.section "HuSample_StopSimple" force
HuSample_StopSimple:
	PHY
	CLY
	jsr HuSample_Stop
	PLY
	rts
.ends




;****************************************************************
; HuSample_Play:						*
;	Play the PCM stream whose number is in A, on the	*
; channel whose number is in Y, using the stream descriptors	*
; at HuSample_DescBase. The channel MUST be placed in DDA mode	*
; and the required volume and panning set beforehand.		*
;****************************************************************
.section "HuSample_Play" force
HuSample_Play:
	CMP HuSample_NumSamples
	beq _outOfRange
	bcc _startPlayback
_outOfRange
	CLA
	STA HuSample_Priority,Y
	rts

_startPlayback
	PHX
	sei

	LDX.w HuSound_HasLFSR,Y	; If the channel was last using LFSR
	beq _skipNoiseKill	; mode, we need to disable it before
				; starting sample playback.
	STY PSG_CHANNEL
	STZ PSG_LFSR
_skipNoiseKill
	TAX			; Our target descriptor should reside
	LDA.w NibSwap_Table,X	; at the descriptor base + (A << 3).
	LSR
	AND #%00000111
	STA HuSample_CurBase + HI
	LDA.w NibSwap_Table,X	; <- No idea why WLA-DX keeps assuming
	ROR			; zp,X instead of abs,X
	AND #%11111000
	STA HuSample_CurBase + LO

	clc
	LDA HuSample_DescBase + LO
	ADC HuSample_CurBase + LO
	STA HuSample_CurBase + LO
	LDA HuSample_DescBase + HI
	ADC HuSample_CurBase + HI
	STA HuSample_CurBase + HI

	TMA #MPR4		; <- We're not using the MCGenjin select macro
	PHA			; here since the playback routine could be called
	LDA MCG_LastMap		; from within the audio update IRQ.
	PHA			; ~! MCGENJIN OBJ/LO SAVE !~

	PHY
	LDA HuSample_DirBase + MCMAP
	LDY SysRegion
	beq _setDirect
	TAY
	LDA MCG_SwapTable,Y
_setDirect
	PLY
	STA MCGENJIN_BANK
	LDA HuSample_DirBase + BANK
	TAM #MPR4		; ~! MCGENJIN OBJ/LO SELECT !~



	LDX #7			; Ensure playback on the target channel
	SXY			; is disabled before we copy in the new 
	STZ HuSample_Playing,X	; sample's information

	LDA (HuSample_CurBase),Y
	STA HuSample_Rem_Hi,X
	DEY
	LDA (HuSample_CurBase),Y
	STA HuSample_Rem_Md,X
	DEY
	LDA (HuSample_CurBase),Y
	STA HuSample_Rem_Lo,X	; Copy the size directly at first...

	bne _noAdjustLen	; ...but if its low byte is zero, then
	LDA HuSample_Rem_Md,X	; knock off one count from the mid and
	DEA			; high bytes to prevent the DEC only
	STA HuSample_Rem_Md,X	; playback routine from missing a page.
	CMP #$FF
	bne _noAdjustLen
	DEC HuSample_Rem_Hi,X
_noAdjustLen

	clc			; Calculate the stream's starting address:
	LDY #1			; Directory Base + Offset
	LDA (HuSample_CurBase),Y
	ADC HuSample_DirBase + LO
	STA HuSample_Src_Lo,X
	INY
	LDA (HuSample_CurBase),Y
	ADC HuSample_DirBase + HI
	CMP #>ADDR_OBJ_HI
	bcc _writeHiAddress
	SBC #>(ADDR_OBJ_HI - ADDR_OBJ_LO)
	sec
_writeHiAddress
	STA HuSample_Src_Hi,X
	INY
	LDA (HuSample_CurBase),Y
	ADC HuSample_DirBase + BANK
	CMP #(BANK_ROM + MCBANK_PAGES)
	bcc _writeBank
	SBC #MCBANK_PAGES
	sec
_writeBank
	STA HuSample_Src_Bank,X
	INY
	LDA (HuSample_CurBase),Y
	ADC HuSample_DirBase + MCMAP
	STA HuSample_Src_McMap,X

	LDA #$FF
	STA HuSound_noWriteback,X
	STA HuSample_Playing,X	; We're all set- disable writebacks,
	STA TIMER_EN		; flag the sample as active, and enable
				; the timer IRQ if it hasn't been already.

	MCG_Res_Low		; =! MCGENJIN OBJ/LO RESTORE !=
	cli
	PLX
	rts
.ends
;****************************************************************
; HuSample_Stop:						*
;	Stop PCM Playback (if active) on the channel in Y.	*
;****************************************************************
.section "HuSample_Stop" force
HuSample_Stop:
	LDA HuSample_Playing + ADDR_RAM,Y
	bne _isActive
	rts
_isActive
	CLA
	sei
	STA HuSample_Playing + ADDR_RAM,Y
	STA HuSound_noWriteback,Y
	STA HuSample_Priority,Y
	cli
	rts
.ends




;****************************************************************
; HuSample_Init:						*
;	Initialize the PCM playback software, and set up the	*
; sample directory at HuSample_DirBase.				*
;****************************************************************
.section "HuSample_Init" force
HuSample_Init:
	STZ HuSample_NumSamples	; Default to zero samples (empty)
	STZ TIMER_EN		; and set the timer to 6.992KHz.
	STZ TIMER_BACKUP

	LDA HuSample_DirBase + MCMAP
	ORA HuSample_DirBase + BANK
	ORA HuSample_DirBase + HI
	ORA HuSample_DirBase + LO
	bne _checkDir		; If the directory base is set to an
	rts			; address other than zero we must
_checkDir			; investigate it in full.
	LDA HuSample_DirBase + MCMAP
	STA MCG_Map
	LDA HuSample_DirBase + BANK
	STA MCG_Bank
	LDA HuSample_DirBase + HI
	STA HuSample_CurStream + HI
	LDA HuSample_DirBase + LO
	STA HuSample_CurStream + LO
	MCG_Sel_Low		; ~! MCGENJIN OBJ/LO SELECT !~

	CLY
_checkHeader
	LDA (HuSample_CurStream),Y
	CMP _validHeader,Y
	bne _done		; Make sure the first three bytes of
	INY			; the directory match the contents
	CPY #3			; of _validHeader. If they check out
	bne _checkHeader	; then copy the sample count.

	LDA (HuSample_CurStream),Y
	STA HuSample_NumSamples
				; Also setup the base address for all
	clc			; the PCM descriptor entries.
	LDA HuSample_DirBase + LO
	ADC #4
	STA HuSample_DescBase + LO
	LDA HuSample_DirBase + HI
	ADC #0
	STA HuSample_DescBase + HI

	LDA HuSample_DirBase + BANK
	CMP #BANK_ROM		; If the directory base starts in the
	bcc _shoveBank		; fixed MCGenjin ROM bank, shove the offset
	bra _done		; into the selectable bank.
_shoveBank
	ADC #BANK_ROM
	STA HuSample_DirBase + BANK
_done
	MCG_Res_Low		; =! MCGENJIN OBJ/LO RESTORE !=
	rts

_validHeader	.db "PCM"
.ends
